#include <sys/wait.h>
#include <assert.h>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <string>
#include <cerrno>
#include <ucontext.h>
#include <iostream>
#include <boost/asio.hpp>

using namespace std;

const int TAILLE_BUFF = 100;

// #######################################################
// Fonction pour gérer les signaux - slide 42 à 45 + TD1
// #######################################################
void     ma_fonction (int num_signal){
  cerr <<  "Le fils " << getpid() << " : Le père me prévient d'un problème" << endl;
}


int
main(int argc, char *argv[])
{
  int pipefd[2];
  pid_t code;
  sighandler_t ancienne_action;

  // #########################################################
  // On test s'il y a un argument sur la ligne de commande pour avoir un message à échanger
  // #########################################################
  //assert(argc == 2);
  if (argc != 2) {
    cerr <<  "Usage " << argv[0] <<  "\"un_truc\" pour faire passer le message 'un truc'"
	 << endl << "  ou " << argv[0] << " erreur pour simuler une erreur de calcul " << endl;
    exit(1);
  }

  // #########################################################
  // Création du pipe 
  // #########################################################
  if (pipe(pipefd) == -1) {
    fprintf(stderr, "erreur du pipe : %s", strerror(errno));
    exit(EXIT_FAILURE);
  }

  // #########################################################
  // Duplication du processus slide 13 à 16
  // #########################################################
  code = fork();
  if (code == -1) {
    perror("fork");
    exit(EXIT_FAILURE);
  }

  
  if (code == 0) {
    // #########################################################
    // #########################################################
    // Code du fils 
    // #########################################################
    // #########################################################
    // Ferme l’extrémité d’écriture inutilisée
    close(pipefd[1]);
    
    //    sleep(2);
    cout << "Le fils " << getpid() << " : j'attend\n" << endl;

    // #########################################################
    // Gestion simple d'un signal pour le fils
    // Récupération du signal SIGUSR1 si le père a un problème
    // #########################################################
    ancienne_action = signal(SIGUSR1, ma_fonction); 
    if (ancienne_action == SIG_ERR) {
      perror("problème avec le gestionaire de signaux ");
      exit(1);
    }

    
    // #########################################################
    // Le fils attend le père et lit le résultat
    // #########################################################
    // creation d'un stream boost::asio à partir du pipe pour assurer l'envoie de données
    boost::asio::io_service service;
    boost::asio::posix::stream_descriptor streampipe(service);    
    streampipe.assign(pipefd[0]);

    // lecture du stream sur au moins 100 octets
    char buf[TAILLE_BUFF];
    boost::system::error_code error;
    size_t len = streampipe.read_some(boost::asio::buffer(buf, TAILLE_BUFF), error);    
    if (error) {
      if (error == boost::asio::error::eof) {
	cerr << "Le fils " << getpid() << " : Le pipe a été fermé par l'autre processus" << endl;
      } else {
	cerr << "Le fils " << getpid() << " : problème à la lecture car " << error.message() << endl;
      }
      exit (1);
    }
    // transformation des octets lus en string
    string mess(buf, len);

    
    cout << "le fils  " << getpid() << " : j'ai reçut : \"" << mess << "\"" << endl;
    

    // On remet en place l'ancien gestionaire de signaux
    ancienne_action = signal(SIGUSR1, ancienne_action);
    if (ancienne_action == SIG_ERR) {
      perror("problème avec le gestionaire de signaux ");
      exit(1);
    }

    // #########################################################
    // Calcul du fils
    // #########################################################
    cout << "Le fils " << getpid() << " : je calcule" << endl;
    sleep(1); // simulation du calcul par un sleep


    // #########################################################
    // terminaison du fils - slide 17 à 21
    // #########################################################
    cout << "Le fils " << getpid() << " : j'ai fini" << endl;
    close(pipefd[0]);
    exit(23);

  } else {
    // #########################################################
    // #########################################################
    // Code du père
    // #########################################################
    // #########################################################
    int status;
    // Ferme l’extrémité de lecture inutilisée
    close(pipefd[0]);
    
    
    // #########################################################
    // Calcul du père
    // #########################################################
    cout << "Le pere " << getpid() << " : je calcule" << endl;
    sleep(1); // simulation du calcul par un sleep

    
    // #########################################################
    // Simulation d'une erreur dans le calcul (qui demande de prévenir le fils)
    // #########################################################
    string mess = argv[1];
    if (mess == "erreur") {
      // Le mot erreur a été passé en argument je souhaite donc simuler un problème de calcul
      cout << "Le pere " << getpid() << " : j'ai un souci, je préviens le fils" << endl;
      kill(code, SIGUSR1); // pour prévenir le fils
      sleep(1); // Pour passer la main au fils.
    } else {
      // #########################################################
      // Le père envoie un méssage au fils
      // #########################################################
      // Conversion du file descriptor en stream boost::asio
      boost::asio::io_service service;
      boost::asio::posix::stream_descriptor streampipe(service, pipefd[1]);
      //      streampipe.assign(pipefd[1]);

      boost::system::error_code error;
      string mess = argv[1];
      cout << "Le pere " << getpid() << " : j'envoie au fils le message \"" << mess << "\"" << endl;
      mess = mess;
      // utilisation de la fonction qui écrit tout ce qu'on lui demande
      write(streampipe, boost::asio::buffer(mess.c_str(), mess.size()),error);
      if (error) {
	cerr << "Le pere " << getpid() << " : impossible d'écrire dans le pipe car " << error.message() << endl;
	exit(1);
      }
    }

    // #########################################################
    // Le père a fini 
    // #########################################################
    cout << "Le pere " << getpid() << " : envoie fini" << endl;
    close(pipefd[1]);       /* Le lecteur verra EOF => le read aura 0 comme résultat*/

    // #########################################################
    // Enterrement du fils - slide 17 à 21
    // #########################################################
    cout << "Le pere " << getpid() << " : j'attend la fin du fils" << endl;
    waitpid(-1,&status, 0);             /* Attente du fils */
    if (WIFEXITED(status)) {
      cout << "Le Père " << getpid() << " : le fils s'est terminé normalement et a renvoyé " << WEXITSTATUS(status) << endl;
    } else {
      cout << "Le Père " << getpid() << " : le fils ne s'est pas terminé normalement" << endl;
    }

    // #########################################################
    // Fin du père
    // #########################################################
    cout << "Le pere " << getpid() << " : j'ai fini" << endl;
    exit(EXIT_SUCCESS);
  }
}
